!! Copyright (C) 2020 M. Oliveira, Heiko Appel
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
!! 02110-1301, USA.
!!

module interaction_surrogate_oct_m
  implicit none

  private
  public ::                &
    interaction_surrogate_t

  !> @brief surrogate interaction class to avoid circular dependencies between
  !> modules.
  !!
  !! This class is used only for the purpose of avoiding circular dependencies
  !! between the interaction_partner_oct_m and interaction_oct_m modules.
  !! Without this surrogate, several methods of interaction_partner_t would have
  !! to take an instance of interaction_t as argument, and therefore
  !! interaction_partner_oct_m would need to use interaction_oct_m. On the other
  !! hand, interaction_t has a pointer to an instance of interaction_partner_t,
  !! so interaction_partner_oct_m would need to use interaction_oct_m.
  !!
  !! NB: This class should only ever be used when strickly necessary. Currently
  !! that is only in the deferred methods of interaction_partner_t that require
  !! an interaction.
  type, abstract :: interaction_surrogate_t
  end type interaction_surrogate_t

end module interaction_surrogate_oct_m

!! Local Variables:
!! mode: f90
!! coding: utf-8
!! End:
